#include "board.h"
#include "search.h"
#include <iostream>
#include "move.h"
#include "input.h"
#include "utils.h"
#include "prophet.h"
#include "stats.h"
#include "killer.h"
#include "hash.h"
#include "hashlist.h"
#include <string.h>
#include "contra.h"
#include "rootmove.h"
#include "book.h"
#include <stdio.h>
#include <fstream>
#include "string.h"


using namespace std;

extern int move_list[];
extern int last_move[];


void PrintMoveList(int* move_list, int n_moves, int wtm);

#include "tree.h"

static float Percent(int total, int fraction){
	return (total > 0) ? ((fraction * 100.0f)/(float)total) : 0;
}


inline void PrintValue(int value){
	if (value == INF)
		cout << "INF";
	else if (value == -INF)
		cout << "-INF";
	else if (value == MateScore(1))
		cout << "MATE";
	else if (value == MateScore(0))
		cout << "-MATE";
	else 
		cout << value;
}


int* pv_move;
int pv_moves_left;

ofstream of;
char* outfile = NULL;


int SearchRoot(int alpha, int beta, int depth, int wtm);
void InitRootMoves(int wtm);

static struct {
	int t0;
	int result;
	int ply;
	int move;
	int best_move;
} analyse_vars;


void PrintBeforeAnalyse(){
	if (outfile && !input_flags.epd_mode){	
		of << board.PrintEPD() << endl;
		of << board.PrintBuf();
	}
	//board.Print();
}


void PrintAfterAnalyse(int wtm){
	int etime = GetTime() - analyse_vars.t0;

	if (input_flags.epd_mode && outfile) {
		of << board.PrintEPD();
		of << " acd " << analyse_vars.ply << "; acn " << stats.nodes << "; acs " << etime/100;
		if (analyse_vars.best_move) of << "; bm "; of << MoveToStr(analyse_vars.best_move);
		of << "; ce " << analyse_vars.result/10 << "; pv ";
		PrintMoveList(root_info.m[0].pv, root_info.m[0].pvlen, wtm);
		of << ";" << endl;
	} else {

	if (outfile)
		of << "Time to analyse position : " << (etime / 100.0f) << " seconds" << endl;
	}

	cout << "Time to analyse position : " << (etime / 100.0f) << " seconds" << endl;
}

void PrintRootPV(){
	Assert(root_info.n_moves > 0);

	cout << analyse_vars.ply << ".	" << analyse_vars.result/10 <<	"	" << (GetTime()-analyse_vars.t0)/100 << "	" << stats.nodes <<  "	" << stats.qnodes  << "	";

	if (input_flags.move_to_ponder) PrintMove(input_flags.move_to_ponder);

	for (int q = 0; q < root_info.m[0].pvlen; q++)
		PrintMove(root_info.m[0].pv[q]);
	cout << endl;

	if (!input_flags.epd_mode) {
		
		if (outfile){
			of << analyse_vars.ply << ".	" << analyse_vars.result/10 <<	"	" << (GetTime()-analyse_vars.t0)/100 << "	" << stats.nodes  << "	" << stats.qnodes << "	";
			for (int q = 0; q < root_info.m[0].pvlen; q++)
				of << MoveToStr(root_info.m[0].pv[q]);
			of << endl;
		} 
			
	}
}

int Board::Analyse(int wtm, int& next_possible_move, int max_ply, int best_move){

	if (BitCount(all) == 2) return 0; // draw : only kings left
	
	ResetCounterMoves();

	cout << "epd diagram : " << board.PrintEPD() << endl;

	//!!!ClearHT();
	//ClearST();

	next_possible_move = 0;

	analyse_vars.best_move = best_move;

	// Probe Opening Book : if a good move is found there it is returned
	if (input_flags.use_books){
		int book_move = ProbeBook(wtm);

		if (book_move) return book_move;
	}


	// Initialization stuff : init piecetables, clear killers
	BuildPiecetable();

	board.EvalPiecetable();
	
	ClearAllKillers();

	stats.nodes = stats.qnodes = 0;

	PrintBeforeAnalyse();	


	analyse_vars.move = 0;
	analyse_vars.t0 = GetTime();
	input_flags.time0 = analyse_vars.t0;

	input_flags.t0 = analyse_vars.t0;

	if (input_flags.time_left <=  50)
		input_flags.time_left = 20; // 0.2 seconds

	root_info.Init(wtm);

	switch(root_info.n_moves){
	case 0 :
		return 0;
		break;
	case 1 :
		if (!input_flags.analyze_mode) 
			return root_info.m[0].move;
		break;
	default:
		;
	}

	if (input_flags.limit_depth > 2){
		max_ply = input_flags.limit_depth;
		input_flags.time_left = 1000000000; // big constant
	}

	if (input_flags.limit_time)
		input_flags.time_left = input_flags.limit_time;

	for (analyse_vars.ply = 2; analyse_vars.ply < max_ply && input_flags.time_left > 0; analyse_vars.ply+= 1){
		/*
		cout << "iteration : " << analyse_vars.ply - 1 << " time left = " << input_flags.time_left;
		cout << " time so far : " << GetTime() - analyse_vars.t0;
		cout << " summed : " << input_flags.time_left + (GetTime() - analyse_vars.t0);
		cout << endl;
		*/

		pvlen[0] = 0;

		int x = SearchRoot(-INF, INF, analyse_vars.ply*PLY_DEPTH, wtm);

		analyse_vars.result = root_info.m[0].score;
		analyse_vars.move = root_info.m[0].move;

		PrintRootPV();

		if (input_flags.time_left <= 0) break;

		if (analyse_vars.result <= -MIN_MATE_SCORE || analyse_vars.result >= MIN_MATE_SCORE)
			break;

		if (root_info.singular) break;
	}

	input_flags.search_score = analyse_vars.result;


	PrintAfterAnalyse(wtm);
	
	next_possible_move = (root_info.m[0].pvlen >= 2) ? root_info.m[0].pv[1] : 0;

	if (analyse_vars.move == 0){
		cout << "Problem : zero time !" << endl;
	}

	//cout << "time used to ponder " << (GetTime() - analyse_vars.t0) << endl;

	/*
	if ((GetTime() - analyse_vars.t0) < saved_time - 100)
		cout << "time warning !" << endl;
	*/

	return analyse_vars.move;
}



extern int TestMove(char* move, int wtm); // in pgn.cpp

void Board::AnalyseEPD(char* filename, char* outfile, int maxtime, int maxply, char* descr){
	FILE* fp;

	ofstream notfound_fp;

	char str[4096];
	char buf[128];

	int nodes = 0;
	int qnodes = 0;

	int best_move_to_find;

	input_flags.analyze_mode = 1;


	stats.Reset();
	
	fp = fopen(filename, "rt");

	if (!fp) {
		cout << "Cannot open file " << filename << endl;
		return;
	}

	notfound_fp.open("notfound.epd");

	if (outfile){
		of.open(outfile);
		if (!of.is_open()){
			cout << "Cannot open output file " << outfile;
			outfile = NULL;
		}
	}

	if (outfile && !input_flags.epd_mode)
		of << "Description : " << descr << endl << endl;

	::outfile = outfile;

	while (!feof(fp)){
		char* result = fgets(str,4096,fp);
		if (result == NULL || strlen(str) < 5) break;

		if (outfile && !input_flags.epd_mode)
			of << "Analysing position : " << str;

		cout << "Analysing position : " << str;
		board.Init(str);
		//board.Print();

		HashListReset();

		best_move_to_find = 0;

		char* bm_str = strstr(str, "bm ");
		if (bm_str){

			bm_str += 3;

			stats.bm++;

			char* p_move = buf;
			for ( ; *bm_str && *bm_str != ';';)
				*p_move++ = *bm_str++;
			*p_move++ = 0;

			cout << "Best move converted " << buf << endl;

			best_move_to_find = TestMove(buf, board.side_to_move);
			if (best_move_to_find == 0){
				stats.bm_not_recognized++;
			} 
		}

		input_flags.time_left = maxtime; // 2 minutes max for each position

		int start_time = GetTime();

		int next_possible_move;
		
		int move = board.Analyse(board.side_to_move, next_possible_move, maxply, best_move_to_find);
		if (move == best_move_to_find){
			stats.bm_found++;
			if (outfile && !input_flags.epd_mode) of << "Best move found !" << endl;
		} else if(best_move_to_find) {
			if (outfile && !input_flags.epd_mode) of << "Best move not found !" << endl;
			if (notfound_fp.is_open())
				notfound_fp << str;
		}

		int stop_time = GetTime();

		nodes += stats.nodes;
		qnodes += stats.qnodes;

		stats.positions++;

		stats.nodes += stats.nodes;
		stats.qnodes += stats.qnodes;

		stats.total_time += stop_time - start_time;

		if (outfile && !input_flags.epd_mode){
			int ph_full, ph_total;
			PawnHTOccupied(ph_full, ph_total);
			of << "pawn hashtable full : " << ph_full << " (" <<  Percent(ph_total, ph_full) << "%)" << endl;
		}
	}

	if (outfile && !input_flags.epd_mode){
		of << endl << endl;

		of << "RESULTS :" << endl << endl;

		int total = nodes + qnodes;

		of << "nodes  = " << nodes << " (" << Percent(total, nodes) << "%)" << endl;
		of << "qnodes = " << qnodes << " (" << Percent(total, qnodes) << "%)" << endl;
		of << "positiones searched = " << stats.positions << endl;
		of << "time taken = " << stats.total_time / 100.0f << " seconds" << endl;
		of << "best moves identified = " << stats.bm << endl;
		of << "  of which were not recognized = " << stats.bm_not_recognized << endl;
		of << "  from the rest, best moves found = " << stats.bm_found << " (" << Percent(stats.bm - stats.bm_not_recognized, stats.bm_found) << "%)" << endl;
		of << "average nps = " << (float)total / (stats.total_time/100.0f) << endl;
		of << "hash probes = " << stats.hash_count << endl;
		of << "hash hits = " << stats.hash_hits << " (" << Percent(stats.hash_count, stats.hash_hits) << "%)" << endl;
		of << "pawn hash replaced = " << stats.pawn_hash_replaced << " (" << Percent(stats.pawn_hash_count, stats.pawn_hash_replaced) << "%)" << endl;
		of << "pawn scores found : " << stats.pawn_score_found << endl;
		of << "pawn scores computed : " << stats.pawn_score_computed << " ";
		of << "("<<Percent(stats.pawn_score_computed+stats.pawn_score_found,stats.pawn_score_computed) << "%)" << endl;
	}

	if (outfile){
		of.close();
		outfile = NULL;
		
	}

	notfound_fp.close();

	fclose(fp);
}


void SortList(int pv[][256], int pvlen[], int score[], int n){
	int done;

	int tmp[256];
	int x;

	done = 0;

	while (!done){
		done = 1;
		for (int i = 0; i+1 <= n; i++)
			if (score[i] < score[i+1]){
				memcpy(tmp, pv[i], sizeof(int)*256);
				memcpy(pv[i], pv[i+1], sizeof(int)*256);
				memcpy(pv[i+1], tmp, sizeof(int)*256);

				x = pvlen[i]; pvlen[i] = pvlen[i+1]; pvlen[i+1] = x;
				x = score[i]; score[i] = score[i+1]; score[i+1] = x;
				done = 0;
			}
	}
}


void Board::AnalyseAllMoves(int wtm, int max_ply, int best_move){
	int local_pv[256][256];
	int local_pv_length[256];

	int local_score[256];

	input_flags.analyze_mode = 1;


	Assert(max_ply >= 3);

	int n;
	
	for (int ply = 2; ply < max_ply; ply++){
		Search(-INF, INF, ply * PLY_DEPTH, 0, wtm, 1);
	}

	int local_mlist[256];
	int* last = board.GenerateAllValidMoves(local_mlist, wtm);

	SaveBoardInfo(0);
	BackupTree(2);

	cout << "Starting searching for solutions on the last ply :" << endl;
	
	n = 0;
	for (int* m = local_mlist; m < last; m++, n++){

		pv_moves_left = 0;
		pv_move = NULL;

		pvlen[0] = 0;


		board.Make(*m, 0, wtm);

		local_score[n] = -Search(-INF, INF, (max_ply-1) * PLY_DEPTH, 0, 1-wtm, 1);
		
		RestoreTree(2);

		board.Unmake(*m, 0, wtm);

		local_pv[n][0] = *m;
		memcpy(local_pv[n]+1, pv[0], pvlen[0] * sizeof(int));
		local_pv_length[n] = pvlen[0]+1;

		cout << n+1 << "(" << last-local_mlist << ").	" << local_score[n] << "	";
		PrintMove(*m);
		for (int i = 0; i < pvlen[0]; i++) PrintMove(pv[0][i]);
		cout << endl;
	}

	SortList(local_pv, local_pv_length, local_score, n-1);

	/*
	cout << "Final move list : " << endl;

	for (int i = 0; i < n; i++){
		cout << i+1 << "(" << last-local_mlist << ").	" << local_score[i]/10 << "	";
		for (int j = 0; j < local_pv_length[i]; j++) PrintMove(local_pv[i][j]);
		cout << endl;
	}
	*/

	if (outfile){
		of << board.PrintBuf();

		of << "Final move list : " << endl;
		
		for (int i = 0; i < n; i++){
			of << i+1 << "(" << last-local_mlist << ").	" << local_score[i]/10 << "	";
			for (int j = 0; j < local_pv_length[i]; j++){
				of << MoveToStr(local_pv[i][j]);
			}
			of << endl;
		}
	}
}



void Board::AnalyseAllMovesEPD(char* filename, char* outfile, int maxtime, int maxply, char* descr){
	FILE* fp;

	input_flags.analyze_mode = 1;

	char str[4096];

	stats.Reset();
	
	fp = fopen(filename, "rt");

	if (!fp) {
		cout << "Cannot open file " << filename << endl;
		return;
	}

	if (outfile){
		of.open(outfile);
		if (!of.is_open()){
			cout << "Cannot open output file " << outfile;
			outfile = NULL;
		}
	}

	if (outfile && !input_flags.epd_mode)
		of << "Description : " << descr << endl << endl;

	::outfile = outfile;

	while (!feof(fp)){
		char* result = fgets(str,4096,fp);
		if (result == NULL || strlen(str) < 5) break;

		if (outfile && !input_flags.epd_mode)
			of << "Analysing position : " << str;

		cout << "Analysing position : " << str;
		board.Init(str);
		board.Print();

		HashListReset();

		input_flags.time_left = maxtime; // xxx minutes max for each position

		int start_time = GetTime();

		AnalyseAllMoves(board.side_to_move, maxply);		

		int stop_time = GetTime();

		stats.positions++;

		stats.nodes += stats.nodes;
		stats.qnodes += stats.qnodes;

		stats.total_time += stop_time - start_time;
	}

	if (outfile){
		of.close();
		outfile = NULL;
	}

	fclose(fp);
}




void Board::CheckMirroredEvalEPD(char* filename, char* outfile){
	FILE* fp;

	input_flags.analyze_mode = 1;

	char str[4096];

	stats.Reset();
	
	fp = fopen(filename, "rt");

	if (!fp) {
		cout << "Cannot open file " << filename << endl;
		return;
	}

	if (outfile){
		of.open(outfile);
		if (!of.is_open()){
			cout << "Cannot open output file " << outfile;
			outfile = NULL;
		}
	}

	::outfile = outfile;

	int matches = 0;
	int nonmatches = 0;

	while (!feof(fp)){
		char* result = fgets(str,4096,fp);
		if (result == NULL || strlen(str) < 5) break;

		board.Init(str);
		
		cout << "[" << matches+nonmatches+1 << "] ";
		of << "[" << matches+nonmatches+1 << "] ";
		
		/*
		cout << "Analysing position : " << str << endl << endl;
		cout << "original board : " << endl;
		board.Print();

		of << "Analysing position : " << str << endl << endl;
		of << "original board : " << endl;
		of << board.PrintBuf() << endl;;
		*/
		
		int eval0 = board.Evaluate(-INF, INF, 0, board.side_to_move);
		
		board.Mirror();
		
		/*
		cout << "mirrored board :" << endl;
		board.Print();

		of << "mirrored board :" << endl;
		of << board.PrintBuf() << endl;;
		*/

		int eval1 = board.Evaluate(-INF, INF, 0, board.side_to_move);

		if (eval1 == eval0){
			/*
			of << " scores match !" << endl;
			cout << " scores match !" << endl;
			*/
			of << "OK";
			cout << "OK";
			matches++;
		} else {
			cout << "NO MATCH " << endl;
			cout << "	score1 = " << eval0 << " score2 = " << eval1 << endl;
			cout << "	position = " << str << endl;
			of << "NO MATCH " << endl;
			of << "	score1 = " << eval0 << " score2 = " << eval1 << endl;
			of << "	position = " << str;
			nonmatches++;
		}

		of << endl;
		cout << endl;
	}

	if (outfile){
		of << "Results : " << endl;
		of << "	" << matches + nonmatches << " results" << endl;
		of << "	" << matches << " matches" << endl;
		of << "	" << nonmatches << " nonmatches" << endl;
	}

	cout << "Results : " << endl;
	cout << "	" << matches + nonmatches << " results" << endl;
	cout << "	" << matches << " matches" << endl;
	cout << "	" << nonmatches << " nonmatches" << endl;

	if (outfile){
		of.close();
		outfile = NULL;
	 
	}

	fclose(fp);
}







void PrintMoveList(int* move_list, int n_moves, int wtm){
	if (outfile){
		for (int i = 0 ;  i < n_moves; i++){
			int to_move = ((i & 1) ? (1-wtm) : wtm);

			if (!board.IsValidMove(move_list[i], to_move)) break;

			board.Make(move_list[i], 0, to_move);

			of << MoveToStrComplete(move_list[i], 1-to_move);
		}
		//of << endl;
	}
}

