#define WIN32_LEAN_AND_MEAN
#define inline __inline
#include <windows.h> // for PeekNamedPipe
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include "debug.h"
#include "board.h"
#include "input.h"
#include "rootmove.h"
#include "move.h"
#include "stats.h"
#include "movelist.h"
#include "hash.h"

using namespace std;

char pause()
{
    char c;
    while ((c = getchar()) != '\n') { }
    return c;
}



#define IsLetter(x) (((x)>='a')&&((x)<='h'))
#define IsDigit(x) (((x)>='1')&&((x)<='8'))

#define     VERSION         "1.068"
#define     BUILDDATE       "25-03-08"

static const DWORD dwTimeout = 0;


InputFlags input_flags;

static int pipe;


const int BUF_SIZE = 256;
static char buf[BUF_SIZE];
static int n_buf = 0;

const int QUEUE_SIZE = 60;

static char queue[QUEUE_SIZE][BUF_SIZE];
static int n_queue = 0;

void ParseCommandLine(char* str);


void Start() {
    cout << "Freyr" << endl;
	cout << "Version 1.068 JA" << endl;

	HANDLE inh = GetStdHandle(STD_INPUT_HANDLE);
	DWORD dw;
	pipe = !GetConsoleMode(inh, &dw);
	if (!pipe)
	{
		SetConsoleMode(inh, dw & ~(ENABLE_MOUSE_INPUT | ENABLE_WINDOW_INPUT));
		FlushConsoleInputBuffer(inh);
 	}
}


void PrintCurrentPV(){
	cout << root_info.m[0].plies << ".	" << root_info.m[0].score << "	" << 0 << "	" << stats.nodes  << "	" << stats.qnodes << "	";
	for (int i = 0; i < root_info.m[0].pvlen; i++)
		PrintMove(root_info.m[0].pv[i]);
	cout << endl;
}

void AddToQueue(){
	cout << "AddTo Queue [" << buf << "]" << endl;

	if (!strcmp(buf, "quit") || !strcmp(buf, "exit")) {
		exit(0);
	}

	if (!strcmp(buf, "force")) {
		input_flags.time_left = -1;
		if (!input_flags.engine_thinking && !input_flags.engine_pondering)
			input_flags.force_mode = 1;
		else
			return;
	}

      
	if (!strcmp(buf, "undo"))
		input_flags.time_left = -1;
	else if (!strcmp(buf, "remove"))
		input_flags.time_left = -1;
	else if (!strcmp(buf, "edit"))
		input_flags.time_left = -1;
	else if (!strcmp(buf, "analyze")){
		input_flags.time_left = -1;
		input_flags.analyze_mode = 1;
	}else if (!strcmp(buf, "black"))
		input_flags.time_left = -1;
	else if (!strcmp(buf, "white"))
		input_flags.time_left = -1;
	if (!strcmp(buf, "time"))
		input_flags.time_left = -1;
	if (!strcmp(buf, "otim"))
		input_flags.time_left = -1;
	
	if (!strcmp(buf, "help"))   {
		printf("\nCommand line options :\n\n\n");
	    printf("Analysing :\n\n");
		printf("-analyse <epd filename>\n");
		printf("-outfile <filename to write results to>\n");
		printf("-maxtime <max time for a position to be analysed in seconds (float)>\n");
        printf("-maxply <max ply to analyse a position>\n");
        printf("-tag <engine configuration description when analysing EPDs>\n");
        printf("-epdmode : epd output\n\n\n");
        printf("Playing:\n\n");
        printf("-xboard\n\n\n");
        printf("press enter to continue\n"), pause();
        printf("General options:\n\n");
        printf("-hashsize <x : integer between 5 and 25> hash size will have 2^x entries\n");
        printf("-tbdir xxx : path to the tablebases files : default tbs/\n\n\n");
        printf("press enter to continue\n"), pause();
        printf("Building an opening book :\n\n");
        printf("it's done in two steps :\n");
        printf("pgn -> temp file -> opening book\n\n");
        printf("example :\n\n");
	    printf("freyr.exe -pgnsource book.pgn -tempfile temp.dat -max_book_moves 15\n");
	    printf("then\n");
	    printf("freyr.exe -tempfile temp.dat -buildbook\n");
	    printf("after which you may delete temp.dat\n\n");
	    printf("Be warned, it's a slow process and takes\n");
	    printf("lots of memory if you want a big book.\n\n");
	    printf("-pgnsouce <name of the file to build the book from>\n");
	    printf("-tempfile <name of the temporary file default temp.dat>\n");
	    printf("-max_book_moves <number of moves to include from each match\n");
        printf(" into the book, default 10>\n");
        printf("-buildbook <name of the temp file>\n\n");
        Start();
        return;
        }
        
        if (!strcmp(buf, "version")) {
		Start();
		return;
	}
        
        if (!strcmp(buf, "about") || !strcmp(buf, "info")) {                      
        printf("\nFreyr 1.068 JA\n\n");
        printf("Release Date : 16/Sep/2000\n\n\n");
        printf("Author : Andrei Fortuna\n");
        printf("andreif@mail.dntis.ro\n");
        printf("http://www.geocities.com/SiliconValley/Lab/3716/\n\n\n");
        printf("A Short Description : Freyr is a chess program written by me\n");
        printf("as a hobby project, Intended as a test ground for various\n");
        printf("chess algorithms. So far it has scored around 1950-2000\n");   
		printf("(AMD K6/2 300 Mhz 64 MB) on FICS. I still have plenty of things\n");
		printf("to try before I'll get it in the 2200+ area.\n\n\n");
		printf("press enter to continue\n"), pause();
		printf("\nInstall notes :\n\n\n");
		printf("copy freyr.exe to your winboard directory and whether\n");
		printf("start winboard with : winboard /fcp freyr.exe\n");
		printf("or enter freyr.exe as the first chess program in\n");
		printf("winboard startup menu\n\n\n");
		printf("You can use Freyr from the command line to analyze\n");
		printf("an epd file, see below.\n\n\n");
		printf("press enter to continue\n"), pause();
		printf("To use an opening book, you need either to download\n");
		printf("a small book from my page or build one from a pgn file,\n");
		printf("see below how. In either case, copy afterwards\n");
		printf("the resulting index.dat and book.bin to the winboard directory\n\n");
		Start();	
		return;
	}
           
	
	if (!strcmp(buf, "?")) {
		input_flags.time_left = -1;
		return;
	}
	
	if (!strcmp(buf, ".") && input_flags.engine_thinking && !input_flags.analyze_mode){
		PrintCurrentPV();
		return;
	}
	
	if (input_flags.engine_thinking && (input_flags.analyze_mode || input_flags.engine_pondering)){ // break analysis if a move is made
		char* str = buf;
		if (IsLetter(str[0]) && IsDigit(str[1]) && IsLetter(str[2]) && IsDigit(str[3]))
			input_flags.time_left = -1;
	}

	Assert(n_queue < QUEUE_SIZE);
	memcpy(queue[n_queue++], buf, BUF_SIZE);
}

void DeleteFirstQueueEntry(){
	for (int i = 0; i+1 < n_queue; i++)
		memcpy(queue[i], queue[i+1], BUF_SIZE);
	n_queue--;
}

int ProcessQueueLine(){
	if (!n_queue) return 0;

	cout << "Process Queue Line [" << queue[0] << "]" << endl;

	static char command[BUF_SIZE];
	memcpy(command, queue[0], BUF_SIZE);
	DeleteFirstQueueEntry();
	ParseCommandLine(command);

	return 1;
}

//-analyse wac.epd -outfile wac.txt -maxtime 5 -maxply 55 -tag "5 seconds"

void PrintStats(){
	cout << "wtm = " << input_flags.wtm;
}


#define ProcessChar(ch) 			\
	if (ch == 8) {					\
		if (n_buf) n_buf--;			\
		return;						\
	}								\
									\
	Assert(n_buf < BUF_SIZE);		\
									\
	buf[n_buf++] = (char)ch;		\
									\
	if (ch == 10 || ch == 13){		\
		buf[--n_buf] = 0;			\
		AddToQueue();				\
		n_buf = 0;					\
	} 


void CheckCommandLine(){
	unsigned long dw;
	int ch;

	if (pipe) {
		if (!PeekNamedPipe(GetStdHandle(STD_INPUT_HANDLE), NULL, 0, NULL, &dw, NULL))
			return;
		for (unsigned int i = 0; i < dw; i++){
			ch = getc(stdin);
			ProcessChar(ch)
		}
	} else {
		if (!kbhit()) return;
		ch = getche();

		ProcessChar(ch);
	}

}

void SetupTime(){
	input_flags.time_left = input_flags.computer_white ? input_flags.white_time : input_flags.black_time;

	//cout << "Setup Time : " << input_flags.white_time << " " << input_flags.black_time << " " << input_flags.time_left << endl;

	// reserve 5 seconds 
	if (input_flags.time_left > 500)
		input_flags.time_left -= 500;
	
	int estimated_moves_left = (MoveNumber() < 25) ? 40 : 28;

	//cout << ">>" << input_flags.time_left << "<<" << endl;

	input_flags.time_left_after = input_flags.time_left;

	input_flags.time_left = input_flags.time_left / estimated_moves_left;

	if (input_flags.level_inc){
		if (input_flags.time_left > (3*input_flags.level_inc + 1)*100){
			input_flags.time_left += input_flags.level_inc *100;
		} 
	} 

	input_flags.time_left_after -= input_flags.time_left;

	//cout << "Setup Time : " << input_flags.level_inc << endl;
	//cout << "Setup Time : " << input_flags.time_left << endl;
	//cout << "time to move : " << input_flags.time_left << endl;
}

void PrintWinboardMove(int move){
	int from = From(move);
	int to = To(move);
	int promote = Promote(move);

	/*
	cout << "1. ... ";
	cout << (char)(File(from)+'a');
	cout << (char)(Rank(from)+'1');
	cout << (char)(File(to)+'a');
	cout << (char)(Rank(to)+'1');

	switch(promote){
	case QUEEN : cout << "=Q"; break;
	case ROOK : cout << "=R"; break;
	case KNIGHT : cout << "=N"; break;
	case BISHOP : cout << "=B"; break;
	}
	*/

	cout << "move ";
	cout << (char)(File(from)+'a');
	cout << (char)(Rank(from)+'1');
	cout << (char)(File(to)+'a');
	cout << (char)(Rank(to)+'1');

	switch(promote){
	case QUEEN : cout << "q"; break;
	case ROOK : cout << "r"; break;
	case KNIGHT : cout << "n"; break;
	case BISHOP : cout << "b"; break;
	}
	
	cout << endl;
}


void EditMode(){
	char c, sq1, sq2;

	static int counter = 0;

	//if (counter++ == 1)
	//while (1) {}

	#define SetSquare {								\
		Assert(IsLetter(sq1));						\
		Assert(IsDigit(sq2)); 						\
		brd[((sq2-'1')<<3)+(sq1-'a')] = (piece_color == WHITE) ? piece : -piece;		\
	}

	#define ClearSquare {							\
		Assert(IsLetter(sq1));						\
		Assert(IsDigit(sq2)); 						\
		brd[((sq2-'1')<<3)+(sq1-'a')] = 0;			\
	}

	cout << "Entering Edit Mode" << endl;

	cout << "The Queue looks like : " << endl;
	for (int x = 0; x < n_queue; x++)
		cout << queue[x] << endl;
	cout << endl;
	
	int brd[64];
	char s[64];
	memset(brd, 0, sizeof(brd));

	int piece_color = WHITE;

	while (1){
		while (n_queue == 0){
			CheckCommandLine();
		}

		memcpy(s, queue[0], 64);
		DeleteFirstQueueEntry();

		cout << "Edit mode : got [" << s << "]" << endl;

		c = s[0];

		//cout << "C = " << c << endl;
		//cout << s << endl;

		if (c == '.') break;

		if (c >= 'A' && c <= 'Z') c = c - 'A' + 'a';

		int piece = 0;
		if (c == 'p'){ piece = PAWN; sq1 = s[1]; sq2 = s[2]; SetSquare}
		else if (c == 'q'){ piece = QUEEN; sq1 = s[1]; sq2 = s[2]; SetSquare}
		else if (c == 'r'){ piece = ROOK; sq1 = s[1]; sq2 = s[2]; SetSquare}
		else if (c == 'n'){ piece = KNIGHT; sq1 = s[1]; sq2 = s[2]; SetSquare}
		else if (c == 'b'){ piece = BISHOP; sq1 = s[1]; sq2 = s[2]; SetSquare}
		else if (c == 'k'){ piece = KING; sq1 = s[1]; sq2 = s[2]; SetSquare}
		else if (c == '#'){ memset(brd, 0, sizeof(brd)); }
		else if (c == 'x'){	sq1 = s[1]; sq2 = s[2]; ClearSquare }
		else if (c == 'c') { piece_color = 1 - piece_color; }
	}

	char str[1024];
	char* p_str = str;

	for (int i = 0; i < 8; i++){
		for (int j = 0; j < 8; j++){
			int sq = ((7-i)<<3)+j;
			switch (brd[sq]){
				case PAWN : *p_str++ = 'P'; break;
				case ROOK : *p_str++ = 'R'; break;
				case KNIGHT : *p_str++ = 'N'; break;
				case BISHOP : *p_str++ = 'B'; break;
				case QUEEN : *p_str++ = 'Q'; break;
				case KING : *p_str++ = 'K'; break;
				case -PAWN : *p_str++ = 'p'; break;
				case -ROOK : *p_str++ = 'r'; break;
				case -KNIGHT : *p_str++ = 'n'; break;
				case -BISHOP : *p_str++ = 'b'; break;
				case -QUEEN : *p_str++ = 'q'; break;
				case -KING : *p_str++ = 'k'; break;
				default : *p_str++ = '1'; break;
			}
		}
		*p_str++ = (i == 7) ? ' ' : '/';
	}

	cout << "EDIT MODE : wtm = " << input_flags.wtm << endl;

	*p_str++ = (input_flags.wtm == WHITE) ? 'w' : 'b';
	*p_str++ = 0;

	cout << "str = " << str << endl;

	//board.Reset();
	board.Init(str);

	ClearHT();
	ClearPawnHT();
	input_flags.white_time = 6000;
	input_flags.black_time = 6000;
	input_flags.force_mode = 0;
	input_flags.time_spent_pondering = 0;
	ResetMoveList(board.side_to_move);

	cout << "Board : ";
	board.Print();
}


void ParseLevelCA(char* str){ // parse level for Chess Academy format
	#define SkipBlank {for (; *str < '0' || *str > '9'; str++) if (*str == 0) return; }

	SkipBlank
	input_flags.level_moves = 0;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_moves = input_flags.level_moves * 10 + (*str-'0');
	}

	// read minutes:seconds
	SkipBlank
	input_flags.level_time = 0;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_time = input_flags.level_time * 10 + (*str-'0');
	}
	SkipBlank
	input_flags.level_time *= 60;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_time = input_flags.level_time * 10 + (*str-'0');
	}

	SkipBlank
	input_flags.level_inc = 0;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_inc = input_flags.level_inc * 10 + (*str-'0');
	}
	#undef SkipBlank

	//cout << "Level [" << input_flags.level_moves << " " << input_flags.level_time << " " << input_flags.level_inc  << "]" << endl;
}


void ParseLevel(char* str){
	if (strstr(str, ":")){
		ParseLevelCA(str);
		return;
	}


	#define SkipBlank {for (; *str < '0' || *str > '9'; str++) if (*str == 0) return; }

	SkipBlank
	input_flags.level_moves = 0;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_moves = input_flags.level_moves * 10 + (*str-'0');
	}

	SkipBlank
	input_flags.level_time = 0;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_time = input_flags.level_time * 10 + (*str-'0');
	}
	input_flags.level_time *= 60;

	SkipBlank
	input_flags.level_inc = 0;
	for (; *str >= '0' && *str <= '9'; *str++){
		if (*str == 0) return;
		input_flags.level_inc = input_flags.level_inc * 10 + (*str-'0');
	}
	#undef SkipBlank

	//cout << "Level [" << input_flags.level_moves << " " << input_flags.level_time << " " << input_flags.level_inc  << "]" << endl;
}

int ParseTime(char* str){
	#define SkipBlank {for (; *str < '0' || *str > '9'; str++) if (*str == 0) return 0; }

	int time = 0;

	SkipBlank

	for (;*str && *str >= '0' && *str <= '9'; *str++){
		time = time * 10 + (*str-'0');
	}

	//cout << "Parse Time : " << time << endl;

	return time;

	#undef SkipBlank
}


void Analyse(){
	int next_possible_move = 0;

	SetupTime();
			
	input_flags.engine_thinking = 1;
	int move = board.Analyse(input_flags.wtm, next_possible_move);
	input_flags.engine_thinking = 0;

	if (input_flags.analyze_mode) // if an analyze command is sent, winboard ignores current move
		return;

	if (move == 0){
		if (Check(input_flags.wtm))
			cout << ((input_flags.wtm) ? "0-1 {Black Mates}" : "1-0 {White Mates}") << endl;
		else
			cout << "1/2-1/2 {Stalemate}" << endl;
		return;
	}

	MakeMove(move);

	input_flags.wtm = 1 - input_flags.wtm;
	PrintWinboardMove(move);
}

void AnalyseHint(){
	int next_possible_move = 0;

	SetupTime();
			
	input_flags.engine_thinking = 1;
	int move = board.Analyse(input_flags.wtm, next_possible_move);
	input_flags.engine_thinking = 0;

	if (move){
		int from = From(move);
		int to = To(move);
		int promote = Promote(move);
		cout << "Hint: ";
		cout << (char)(File(from)+'a');
		cout << (char)(Rank(from)+'1');
		cout << (char)(File(to)+'a');
		cout << (char)(Rank(to)+'1');

		switch(promote){
		case QUEEN : cout << "q"; break;
		case ROOK : cout << "r"; break;
		case KNIGHT : cout << "n"; break;
		case BISHOP : cout << "b"; break;
		}
	
		cout << endl;
	}
}


void AnalyseDraw(){
	if (root_info.n_moves && root_info.m[0].score == 0)
		cout << "I accept" << endl;
	else
		cout << "I decline" << endl;
}


void ParseCommandLine(char* str){
	int dummy;

	if (IsLetter(str[0]) && IsDigit(str[1]) && IsLetter(str[2]) && IsDigit(str[3])){
		cout << "The board looks like this :" << endl;
		board.Print();

		int from = Square(str[1]-'1',str[0]-'a');
		int to = Square(str[3]-'1',str[2]-'a');
		int promote;
		switch (str[4]){
		case 'Q':
		case 'q':	promote = QUEEN; break;
		case 'N':
		case 'n':	promote = KNIGHT; break;
		case 'B':
		case 'b':	promote = BISHOP; break;
		case 'R':
		case 'r':	promote = ROOK; break;
		default :	promote = 0; break;
		}

		MakeMove(from, to, promote);

		input_flags.wtm = 1 - input_flags.wtm;

		if (input_flags.analyze_mode){
			input_flags.engine_thinking = 1;
			input_flags.time_left = 6000000; // plenty
			
			int move = board.Analyse(input_flags.wtm, dummy);
		} else if (!input_flags.force_mode) {
			Analyse();

			if (board.GenerateAllValidMoves(move_list, input_flags.wtm) == move_list){
				if (Check(input_flags.wtm))
					cout << ((input_flags.wtm) ? "0-1 {Black Mates}" : "1-0 {White Mates}") << endl;
				else
					cout << "1/2-1/2 {Stalemate}" << endl;
				return;
			}
			

			if (input_flags.do_pondering){
				int dummy;

				cout << endl;
				cout << "Starting pondering :" << endl << endl;
				board.Print();

				input_flags.time_left = 6000000; // plenty
				input_flags.engine_thinking = 1;
				input_flags.engine_pondering = 1;
				int move = board.Analyse(input_flags.wtm, dummy);
				input_flags.engine_pondering = 0;
				input_flags.engine_thinking = 0;
			}
		}
		return;
	}
	
	 if (!strncmp(str, "protover", 8)){
	          printf("feature colors=1 myname=\"Freyr %s by Andrei Fortuna\"",VERSION);
              printf(" time=1 pause=0 variants=\"normal\" sigint=0 sigterm=0");
              printf(" name=1 playother=1 setboard=1 draw=1 done=1");
     
         }else
            
            if (!strncmp(str,"variant", 7)) {
	    	char variant[100];
	    	scanf("%s",variant);
	    	if (!strcmp(variant,"normal")) {
			printf("Error: unsupported variant.");
			}
      }else
		
		if (!strncmp(str, "result", 6)) {
           char result[256];
           char line[100];
           sscanf(line,"result %s",result);
   }else
  	    
	

	if (!strncmp(str, "level", 5)){
		ParseLevel(str+6);
		input_flags.white_time = input_flags.black_time = input_flags.level_time * 100;
		input_flags.limit_depth = 0;
		input_flags.limit_time = 0;
	} else if (!strcmp(str, "new")){
		ClearHT();
		ClearPawnHT();
		board.Reset();
		input_flags.computer_white = 0;
		input_flags.show_thinking = 1;
		input_flags.wtm = 1;
		input_flags.white_time = 6000;
		input_flags.black_time = 6000;
		input_flags.force_mode = 0;
		input_flags.time_spent_pondering = 0;
		input_flags.engine_pondering = 0;
		input_flags.edited_position = 0;
		input_flags.force_mode = 0;
		ResetMoveList(1);
	} else if (!strncmp(str, "hash", 4)){
		int size = ParseTime(str+4);
		ResizeHT(size);
	} else if (!strncmp(str, "draw", 4)){
		AnalyseDraw();
	} else if (!strcmp(str, "books on")){
		input_flags.use_books = 1;
	} else if (!strcmp(str, "books off")){
		input_flags.use_books = 0;
	} else if (!strncmp(str, "time", 4)){
		int time = ParseTime(str+4);
		input_flags.white_time = time;
		input_flags.black_time = time;
	} else if (!strncmp(str, "sd", 2)){
		int sd = ParseTime(str+2);
		cout << "search depth mode sd = " << sd << endl;
		input_flags.limit_depth = sd;
		input_flags.limit_time = 0;
	} else if (!strncmp(str, "st", 2)){
		int st = ParseTime(str+2);
		cout << "search time mode st = " << st << endl;
		input_flags.limit_depth = 0;
		input_flags.limit_time = st;
	} else if (!strncmp(str, "otim", 4)){
		int time = ParseTime(str+4);;
		if (!input_flags.computer_white)
			input_flags.white_time = time;
		else
			input_flags.black_time = time;
	} else if (!strcmp(str, "post")){
		input_flags.show_thinking = 1;
	} else if (!strcmp(str, "nopost")){
		input_flags.show_thinking = 0;
	} else if (!strcmp(str, "force")){
		input_flags.force_mode = 1;
	} else if (!strcmp(str, "white")){
		input_flags.computer_white = 1;
	} else if (!strcmp(str, "black")){
		input_flags.computer_white = 0;
	} else if (!strcmp(str, "hard")){
		input_flags.do_pondering = 1;
	} else if (!strcmp(str, "easy")){
		input_flags.do_pondering = 0;
	} else if (!strcmp(str, "go")){
		input_flags.force_mode = 0;
		if (input_flags.wtm == input_flags.computer_white){	// start thinking
			Analyse();
		}
	} else if (!strcmp(str, "hint")){
		input_flags.force_mode = 0;
		AnalyseHint();
	} else if (!strcmp(str, "edit")){
		EditMode();
		input_flags.edited_position = 1;
		if (input_flags.analyze_mode) {
			input_flags.time_left = 6000000;
			input_flags.engine_thinking = 1;
			int move = board.Analyse(input_flags.wtm, dummy);
			input_flags.engine_thinking = 0;
		}
	} else if (!strcmp(str, "undo")){
		if (MoveNumber() > 0){
			UndoMove();

			input_flags.wtm = 1 - input_flags.wtm;
			
			if (input_flags.analyze_mode) {
				input_flags.time_left = 6000000;
				input_flags.engine_thinking = 1;
				int move = board.Analyse(input_flags.wtm, dummy);
				input_flags.engine_thinking = 0;
			} 
		}
	} else if (!strcmp(str, "remove")){
		if (MoveNumber() > 1){
			UndoMove();
			UndoMove();
			if (input_flags.analyze_mode) {
				input_flags.time_left = 6000000;
				input_flags.engine_thinking = 1;
				int move = board.Analyse(input_flags.wtm, dummy);
				input_flags.engine_thinking = 0;
			}
		}
	} else if (!strcmp(str, "print")){
		board.Print();
	} else if (!strcmp(str, "analyze")){
		input_flags.analyze_mode = 1;
		input_flags.time_left = 6000000;
		input_flags.engine_thinking = 1;
		int move = board.Analyse(input_flags.wtm, dummy);
		input_flags.engine_thinking = 0;
	} else if (!strcmp(str, "exit")){
		input_flags.analyze_mode = 0;
	} 

}



void InitCommandLine(){
	input_flags.engine_thinking = 0;
	input_flags.analyze_mode = 0;
	input_flags.force_mode = 0;

	input_flags.limit_depth = 0;
	input_flags.limit_time = 0;

	input_flags.use_books = 1;

    Start();

	ParseCommandLine("new");	

	while (1){
		Sleep(100);
		CheckCommandLine();
		while (ProcessQueueLine())
			;
	}
}
