#include "hash.h"
#include "board.h"
#include <string.h>
#include <iostream>
#include "stats.h"

using namespace std;


static int HASH_SIZE = 0;
static int HASH_MASK = 0;
#define VERIFY_HASHSTORE 0

const int PAWN_HASH_SIZE = 1 << 18;
const int PAWN_HASH_MASK = PAWN_HASH_SIZE - 1;

static PawnHashEntry pawn_hashtable[PAWN_HASH_SIZE];



static HashEntry* hashtable = NULL;


#define HT_INDEX ((unsigned int)(board.hashcode & HASH_MASK)<<1)

void InitHT(int size){
	if (size < 17 || size > 25) 
		error("Invalid hash table size", __FILE__, __LINE__);

	cout << "InitHT [" << size << "]" << endl;

	HASH_SIZE = 1 << (size-1);
	HASH_MASK = HASH_SIZE-1;
	hashtable = new HashEntry[HASH_SIZE*2];
	
	ClearHT();
}

void ResizeHT(int x){
	// convert from MB to bytes
	x *= 1024*1024; 

	// substract pawn ht size
	x -= sizeof(pawn_hashtable);

	cout << "pawn hash = " << sizeof(pawn_hashtable) << endl;
	
	cout << "x1 = [" << x << "]" << endl;

	// divide by 2 and by size of the HTEntry and by 2 (2 layers htable)
	x /= sizeof(HashEntry);

	cout << "x2 = [" << x << "]" << endl;

	x /= 2;

	cout << "x3 = [" << x << "]" << endl;

	delete[] hashtable;
	hashtable = NULL;

	if (x < (1 << 17)){
		InitHT(17);
		return;
	}

	for (int size = 17; size < 25; size++)
		if ((1 << size) > x){
			InitHT(size);
			return;
		}

	InitHT(25);
	return;
}



void ClearHT(){
	memset(hashtable, 0, sizeof(HashEntry)*HASH_SIZE*2);
}


const HASHCODE hashentry_hashmask = OneShl(48)-1;
const HASHCODE hashentry_agedepth_mask = Compl(hashentry_hashmask);
const HASHCODE hashentry_age_mask = Shl(OneShl(3)-1, 48);
const HASHCODE hashentry_depth_mask = Shl(OneShl(13)-1, 51);

const int hashentry_move_mask = (1 << 21) - 1;
const int hashentry_flag_mask = ((1 << 2) - 1)<<21;

inline int HashCompare(HashEntry* he){
	return ((he->qword & hashentry_hashmask) == (board.hashcode & hashentry_hashmask));
}

inline unsigned int HashGetDepth(HashEntry* he){
	return (unsigned int)(And(Shr(he->qword,51), OneShl(13)-1));
}

inline unsigned int HashGetAge(HashEntry* he){
	return (unsigned int)(And(Shr(he->qword,48), OneShl(3)-1));
}

inline void HashSetHashcode(HashEntry* he){
	he->qword = (board.hashcode & hashentry_hashmask) | (he->qword & hashentry_agedepth_mask);
}

inline void HashSetDepth(HashEntry* he, unsigned int depth){
	he->qword = (he->qword & (hashentry_hashmask | hashentry_age_mask)) | (Shl(((HASHCODE)depth), 51));
}

inline void HashSetAge(HashEntry* he, unsigned int age){
	he->qword = (he->qword & (hashentry_hashmask | hashentry_depth_mask)) | (Shl(((HASHCODE)age), 48));
}

inline unsigned int HashGetMove(HashEntry* he){
	return he->word1 & hashentry_move_mask;
}

inline unsigned int HashGetFlag(HashEntry* he){
	return (he->word1 >> 21);
}

inline void HashSetMove(HashEntry* he, int move){
	he->word1 = ((he->word1 >> 21) << 21) + move;
}

inline void HashSetFlag(HashEntry* he, int flag){
	he->word1 = (he->word1 & hashentry_move_mask) | (flag << 21);
}


int ProbeHT(int& alpha, int& beta, int depth, int& score, int& hash_move){
	HashEntry* h = hashtable+HT_INDEX;

	stats.hash_count++;

	hash_move = 0;

	if (HashGetFlag(h) == UNUSED_HASH)
		return 0;
	else {
		if (!HashCompare(h)){
			h++;
			if (HashGetFlag(h) == UNUSED_HASH || !HashCompare(h))
				return 0;
		} 
		stats.hash_hits++;
	}

	hash_move = HashGetMove(h);

	if (HashGetDepth(h) < depth)
		return 0;

	switch(HashGetFlag(h)){
	case EXACT:
		break;
	case FAIL_HIGH:	
		if (h->score < beta)
			return 0;
		//alpha = h->score;
		break;
	case FAIL_LOW:
		if (h->score > alpha)
			return 0;
		//beta = h->score;
		break;
	default:
		return 0;
	}
	
	score = h->score;

	#if 1
	if (score < alpha) score = alpha;
	if (score > beta) score = beta;
	#endif

	return 1;
}

void AddHT(int score, int flag, int depth, int move){
	HashEntry* h = hashtable+HT_INDEX;

	// h[0] is the depth-prefered, h[1] is always store

	Assert(flag == EXACT || flag == FAIL_LOW || flag == FAIL_HIGH);

	if (HashGetFlag(h) == UNUSED_HASH || HashGetAge(h) != board.age || HashGetDepth(h) < depth){
		
		if (HashGetFlag(h) != UNUSED_HASH && !HashCompare(h)) {
			stats.hash_replacements++;
			// move the previous entry from depth-preffered to always store
			h[1] = h[0];
		}
		
		#if VERIFY_HASHSTORE
			Assert(depth >= 0 && depth < 8192);
			Assert(flag >= 0 && flag < 4);
			Assert(board.age >= 0 && board.age < 8);
		#endif

		HashSetHashcode(h);
		HashSetFlag(h, flag);
		HashSetDepth(h, (unsigned int)depth);
		HashSetMove(h, (unsigned int) move);
		HashSetAge(h, (unsigned int)board.age);
		h->score = score;

		#if VERIFY_HASHSTORE
			Assert(flag == HashGetFlag(h));
			Assert(depth == HashGetDepth(h));
			Assert(move == HashGetMove(h));
			Assert(board.age == HashGetAge(h));
			Assert(HashCompare(h));
		#endif

	} else {
		h++;

		if (HashGetFlag(h) != UNUSED_HASH && !HashCompare(h)) stats.hash_replacements++;

		#if VERIFY_HASHSTORE
			Assert(depth >= 0 && depth < 8192);
			Assert(flag >= 0 && flag < 4);
			Assert(board.age >= 0 && board.age < 8);
		#endif

		HashSetHashcode(h);
		HashSetFlag(h, flag);
		HashSetDepth(h, (unsigned int)depth);
		HashSetMove(h, (unsigned int) move);
		HashSetAge(h, (unsigned int)board.age);
		h->score = score;

		#if VERIFY_HASHSTORE
			Assert(flag == HashGetFlag(h));
			Assert(depth == HashGetDepth(h));
			Assert(move == HashGetMove(h));
			Assert(board.age == HashGetAge(h));
			Assert(HashCompare(h));
		#endif
	}
}




void AddPawnHT(PawnHashEntry& phash){
	PawnHashEntry& he = pawn_hashtable[board.pawn_hashcode & PAWN_HASH_MASK];

	if (he.hashcode || he.pawn_score)
		stats.pawn_hash_replaced++;

	memcpy(&he, &phash, sizeof(PawnHashEntry));

	he.hashcode = board.pawn_hashcode;

	stats.pawn_hash_count++;
}

PawnHashEntry* ProbePawnHT(){
	PawnHashEntry& he = pawn_hashtable[board.pawn_hashcode & PAWN_HASH_MASK];
	if (he.hashcode != board.pawn_hashcode) {
		stats.pawn_score_computed++;
		return NULL;
	}

	stats.pawn_score_found++;
	
	stats.pawn_hash_hits++;

	return &he;	
}

void ClearPawnHT(){
	memset(pawn_hashtable, 0, sizeof(pawn_hashtable));
}

void PawnHTOccupied(int& full, int& total){
	full = 0;
	for (int i = 0; i < PAWN_HASH_SIZE; i++){
		PawnHashEntry& he = pawn_hashtable[i];
		if (he.hashcode != 0 || he.pawn_score != 0)
			full++;
	}
	total = PAWN_HASH_SIZE;
}

