#include "movelist.h"
#include "board.h"
#include "move.h"
#include "hashlist.h"
#include "history.h"
#include "tree.h"
#include <iostream>

using namespace std;

class MoveList{
public:
	int n;
	int list[1024];
	BITBOARD hashcode[1024];
	SaveInfo saved_list[1024];
	int wtm;

	void Reset(int wtm);
	void MakeMove(int from, int to, int promote);
	void MakeMove(int move);
	void UndoMove();

	int IsDraw();
};

static MoveList m_list;

static int tmp_move_list[1024];


void MoveList::Reset(int wtm){
	HashListReset();
	this->wtm = wtm;
	n = 0;
}

void MoveList::MakeMove(int from, int to, int promote){
	int*m;
	int* last = board.GenerateAllMoves(tmp_move_list, wtm);

	SaveBoardInfo_forMove(n);

	#if 0
	/*
	cout << "Make Move : "; cout << endl;
	cout << "Board looks like this : " << endl;
	board.Print();
	cout << endl;

	//cout <<  "HashCode = "; Print(board.hashcode); cout <<endl;
	*/

	cout << "Available moves : ";
	for (m = tmp_move_list; m < last; m++) 
		PrintMove(*m);
	cout << endl;
	#endif
	
	for (m = tmp_move_list; m < last; m++){
		if (From(*m) == from && To(*m) == to && Promote(*m) == promote){
			list[n] = *m;
			hashcode[n] = board.hashcode;

			HashListAdd();

			AgeHistory();

			board.Make(*m, 0, wtm);

			saved_list[n] = save_struct[0];
			wtm = 1 - wtm;
			board.age = (board.age + 1) & 7;
			n++;
			return;
		}
	}
	
	cout << "Make move : cannot find move " << from << " " << to << " " << promote << endl;
}

void MoveList::MakeMove(int move){
	int* last = board.GenerateAllMoves(tmp_move_list, wtm);

	SaveBoardInfo_forMove(n);

	#if 0
	cout << "Make Move : "; PrintMove(move); cout << endl;
	cout << "Board looks like this : " << endl;
	board.Print();
	cout << endl;

	cout <<  "HashCode = "; Print(board.hashcode); cout <<endl;
	#endif

	for (int* m = tmp_move_list; m < last; m++){
		if (*m == move){
			list[n] = *m;
			hashcode[n] = board.hashcode;

			HashListAdd();

			AgeHistory();

			board.Make(*m, 0, wtm);

			saved_list[n] = save_struct[0];
			wtm = 1 - wtm;
			board.age = (board.age + 1) & 7;
			n++;
			return;
		}
	}
	
	#if 0
	cout << "Make move : cannot find move ";
	PrintMove(move);
	cout << endl;
	#endif
}

void MoveList::UndoMove(){

	#if 0
	cout << "UndoMove :  n = " << n << endl;
	cout << "The board before: " << endl;
	board.Print();
	cout <<endl;
	#endif

	if (n <= 0)
		cout << "UndoMove : cannot undo" << endl;
	else {
		RestoreBoardInfo_forMove(n-1);

		wtm = 1 - wtm;
		save_struct[0] = saved_list[n-1];
		board.Unmake(list[n-1], 0, wtm);
		board.age = (board.age + 8 - 1) & 7;
		--n;

		//cout <<  "HashCode = "; Print(board.hashcode); cout <<endl;

		HashListRemoveLast();
	}
	
	#if 0
	cout << "The board after: " << endl;
	board.Print();
	cout <<endl;
	#endif
}

int MoveList::IsDraw(){	// detect 50 moves no capture or 3 times same position draws
	if (n > 100){
		int draw = 1;
		for (int i = n - 100; i < n; i++)
			if (Captured(list[i]) || Piece(list[i]) == PAWN){ draw = 0; break; }
		if (draw) return 1;
	}	

	if (n > 1){
		int count = 1;
		for (int i = 0; i+1 < n; i++)
			if (hashcode[i] == hashcode[n-1])
				count++;
		if (count >= 3) return 1;
	}

	return 0;
}



void ResetMoveList(int wtm){
	m_list.Reset(wtm);
}

void MakeMove(int from, int to, int promote){
	m_list.MakeMove(from, to, promote);
}	

void MakeMove(int move){
	m_list.MakeMove(move);
}

int Rule50Draw(){
	return m_list.IsDraw();
}

void UndoMove(){
	m_list.UndoMove();
}


int MoveNumber(){
	return m_list.n;
}


void GetMoveList(int** p, int& n){
	*p = m_list.list;
	n = m_list.n;
}
