#include "board.h"
#include <iostream>
#include "swap.h"
#include "move.h"

using namespace std;

#define VERIFY_CHECK_EVASIONS 0

int* Board::GenerateCaptures(int* move, int wtm){
	register BITBOARD piecebd, moves, promotions, pcapturesl, pcapturesr, target;
	register int from, to, temp;

	if (wtm){
		// knight moves
		piecebd = wKnights;
		while (piecebd){
			from = FirstOne(piecebd);
			moves = And(knight_attacks[from], black_pieces);
			temp = from + (KNIGHT << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|((-PieceOnSquare(to)) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// rook moves
		piecebd = wRooks;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksRook(from), black_pieces);
			temp = from + (ROOK << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|((-PieceOnSquare(to)) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// bishop moves
		piecebd = wBishops;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksBishop(from), black_pieces);
			temp = from + (BISHOP << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|((-PieceOnSquare(to)) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// queen moves
		piecebd = wQueens;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksQueen(from), black_pieces);
			temp = from + (QUEEN << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|((-PieceOnSquare(to)) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// king moves
		from = white_king_sq;
		moves = And(king_attacks[from], black_pieces);
		temp = from + (KING << 12);
		while (moves){
			to = FirstOne(moves);
			*move++ = temp|(to << 6)|((-PieceOnSquare(to)) << 15);
			ClearFirstOne(moves);
		}

		// pawn moves - first 1 square forward promotions, then ep, then captures
		promotions = And(Shl(And(wPawns, rank_mask[RANK7]),8),Compl(all));
		while (promotions){
			to = FirstOne(promotions);
			*move++ = (to-8)|(to<<6)|(PAWN<<12)|(QUEEN<<18);
			ClearFirstOne(promotions);
		}

		target = ep_square ? Or(black_pieces, set_mask[ep_square]) : black_pieces;
		pcapturesl = And(Shl(And(wPawns, mask_left_edge),7),target);
		while (pcapturesl){
			to = FirstOne(pcapturesl);
			if (to < 56){
				if (PieceOnSquare(to))
					*move++ = (to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15);
				else // en passant
					*move++ = (to-7)|(to<<6)|(PAWN<<12)|(PAWN<<15);
			} else // promotion
				*move++ = (to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(QUEEN<<18);
			ClearFirstOne(pcapturesl);
		}

		pcapturesr = And(Shl(And(wPawns, mask_right_edge),9),target);
		while (pcapturesr){
			to = FirstOne(pcapturesr);
			if (to < 56){
				if (PieceOnSquare(to))
					*move++ = (to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15);
				else // en passant
					*move++ = (to-9)|(to<<6)|(PAWN<<12)|(PAWN<<15);
			} else // promotion
				*move++ = (to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(QUEEN<<18);
			ClearFirstOne(pcapturesr);
		}
	
	
	} else { // black to move

		// knight moves
		piecebd = bKnights;
		while (piecebd){
			from = FirstOne(piecebd);
			moves = And(knight_attacks[from], white_pieces);
			temp = from + (KNIGHT << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|(PieceOnSquare(to) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// rook moves
		piecebd = bRooks;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksRook(from), white_pieces);
			temp = from + (ROOK << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|(PieceOnSquare(to) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// bishop moves
		piecebd = bBishops;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksBishop(from), white_pieces);
			temp = from + (BISHOP << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|(PieceOnSquare(to) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// queen moves
		piecebd = bQueens;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksQueen(from), white_pieces);
			temp = from + (QUEEN << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6)|(PieceOnSquare(to) << 15);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// king moves
		from = black_king_sq;
		moves = And(king_attacks[from], white_pieces);
		temp = from + (KING << 12);
		while (moves){
			to = FirstOne(moves);
			*move++ = temp|(to << 6)|(PieceOnSquare(to) << 15);
			ClearFirstOne(moves);
		}

		// pawn moves - first 1 square forward promotions, then ep, then captures
		promotions = And(Shr(And(bPawns, rank_mask[RANK2]),8),Compl(all));
		while (promotions){
			to = FirstOne(promotions);
			*move++ = (to+8)|(to<<6)|(PAWN<<12)|(QUEEN<<18);
			ClearFirstOne(promotions);
		}

		target = ep_square ? Or(white_pieces, set_mask[ep_square]) : white_pieces;
		pcapturesl = And(Shr(And(bPawns, mask_left_edge),9),target);
		while (pcapturesl){
			to = FirstOne(pcapturesl);
			if (to > 7){
				if (PieceOnSquare(to))
					*move++ = (to+9)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15);
				else // en passant
					*move++ = (to+9)|(to<<6)|(PAWN<<12)|(PAWN<<15);
			} else  // promotion
				*move++ = (to+9)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(QUEEN<<18);
			ClearFirstOne(pcapturesl);
		}

		pcapturesr = And(Shr(And(bPawns, mask_right_edge),7),target);
		while (pcapturesr){
			to = FirstOne(pcapturesr);
			if (to > 7){
				if (PieceOnSquare(to))
					*move++ = (to+7)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15);
				else // en passant
					*move++ = (to+7)|(to<<6)|(PAWN<<12)|(PAWN<<15);
			} else // promotion
				*move++ = (to+7)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(QUEEN<<18);
			ClearFirstOne(pcapturesr);
		}

	}

	return move;
}










int* Board::GenerateNonCaptures(int* move, int wtm){
	register BITBOARD piecebd, moves, pcapturesl, pcapturesr, target;
	register BITBOARD padvances1, padvances2, promotions;
	register int from, to, temp;

	if (wtm){

		// first castle moves
		if (castle_flags & (WHITE_LCASTLE | WHITE_SCASTLE)){
			if (Not(And(white_scastle_mask, all)))
			if ((castle_flags & WHITE_SCASTLE) && !Attacked(E1, 0) && !Attacked(F1, 0) && !Attacked(G1, 0))
				*move++ = (E1)|(G1 << 6)|(KING<<12);
			if (Not(And(white_lcastle_mask, all)))
			if ((castle_flags & WHITE_LCASTLE) && !Attacked(C1, 0) && !Attacked(D1, 0) && !Attacked(E1, 0))
				*move++ = (E1)|(C1 << 6)|(KING<<12);
		}

		target = Compl(all);

		// knight moves
		piecebd = wKnights;
		while (piecebd){
			from = FirstOne(piecebd);
			moves = And(knight_attacks[from], target);
			temp = from + (KNIGHT << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// rook moves
		piecebd = wRooks;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksRook(from), target);
			temp = from + (ROOK << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// bishop moves
		piecebd = wBishops;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksBishop(from), target);
			temp = from + (BISHOP << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// queen moves
		piecebd = wQueens;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksQueen(from), target);
			temp = from + (QUEEN << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// king moves
		from = white_king_sq;
		moves = And(king_attacks[from], target);
		temp = from + (KING << 12);
		while (moves){
			to = FirstOne(moves);
			*move++ = temp|(to << 6);
			ClearFirstOne(moves);
		}

		// pawn moves - 1 sq fwd, 2 sq fwd, left and right captures on 8th rank
		padvances1 = And(And(Shl(wPawns,8),target), Compl(rank_mask[RANK8]));
		padvances2 = And(Shl(And(padvances1, rank_mask[RANK3]),8),target);

		while (padvances2) {
			to = FirstOne(padvances2);
			*move++ = (to-16)|(to<<6)|(PAWN<<12);
			ClearFirstOne(padvances2);
		}

		while (padvances1) {
			to = FirstOne(padvances1);
			if (to < 56)
				*move++ = (to-8)|(to<<6)|(PAWN<<12);
			else {
				*move++=(to-8)|(to<<6)|(PAWN<<12)|(ROOK<<18);
				*move++=(to-8)|(to<<6)|(PAWN<<12)|(BISHOP<<18);
				*move++=(to-8)|(to<<6)|(PAWN<<12)|(KNIGHT<<18); 				
			}
			ClearFirstOne(padvances1);
		}

		target = And(black_pieces, rank_mask[RANK8]);
		pcapturesl = And(Shl(And(wPawns, mask_left_edge),7),target);
		pcapturesr = And(Shl(And(wPawns, mask_right_edge),9),target);

		while (pcapturesl) {
			to=FirstOne(pcapturesl);
			*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(ROOK<<18);
			*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(BISHOP<<18);
			*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(KNIGHT<<18);
			ClearFirstOne(pcapturesl);
		}
		while (pcapturesr) {
			to=FirstOne(pcapturesr);
			*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(ROOK<<18);
			*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(BISHOP<<18);
			*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(KNIGHT<<18);
			ClearFirstOne(pcapturesr);
		}

		promotions = And(Shl(And(wPawns, rank_mask[RANK7]),8),Compl(all));
		while (promotions){
			to = FirstOne(promotions);
			*move++ = (to-8)|(to<<6)|(PAWN<<12)|(KNIGHT<<18);
			*move++ = (to-8)|(to<<6)|(PAWN<<12)|(ROOK<<18);
			*move++ = (to-8)|(to<<6)|(PAWN<<12)|(BISHOP<<18);
			ClearFirstOne(promotions);
		}


	} else { // black to move

		// first castle moves
		if (castle_flags & (BLACK_LCASTLE | BLACK_SCASTLE)){
			if (Not(And(black_scastle_mask, all)))
			if ((castle_flags & BLACK_SCASTLE) && !Attacked(E8, 1) && !Attacked(F8, 1) && !Attacked(G8, 1))
				*move++ = (E8)|(G8 << 6)|(KING<<12);
			if (Not(And(black_lcastle_mask, all)))
			if ((castle_flags & BLACK_LCASTLE) && !Attacked(C8, 1) && !Attacked(D8, 1) && !Attacked(E8, 1))
				*move++ = (E8)|(C8 << 6)|(KING<<12);
		}

		target = Compl(all);

		// knight moves
		piecebd = bKnights;
		while (piecebd){
			from = FirstOne(piecebd);
			moves = And(knight_attacks[from], target);
			temp = from + (KNIGHT << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// rook moves
		piecebd = bRooks;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksRook(from), target);
			temp = from + (ROOK << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// bishop moves
		piecebd = bBishops;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksBishop(from), target);
			temp = from + (BISHOP << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// queen moves
		piecebd = bQueens;
		while (piecebd){
			from = FirstOne(piecebd);

			moves = And(AttacksQueen(from), target);
			temp = from + (QUEEN << 12);
			while (moves){
				to = FirstOne(moves);
				*move++ = temp|(to << 6);
				ClearFirstOne(moves);
			}
			ClearFirstOne(piecebd);
		}

		// king moves
		from = black_king_sq;
		moves = And(king_attacks[from], target);
		temp = from + (KING << 12);
		while (moves){
			to = FirstOne(moves);
			*move++ = temp|(to << 6);
			ClearFirstOne(moves);
		}

		// pawn moves - 1 sq fwd, 2 sq fwd, left and right captures on 8th rank
		padvances1 = And(And(Shr(bPawns,8),target), Compl(rank_mask[RANK1]));
		padvances2 = And(Shr(And(padvances1, rank_mask[RANK6]),8),target);

		while (padvances2) {
			to = FirstOne(padvances2);
			*move++ = (to+16)|(to<<6)|(PAWN<<12);
			ClearFirstOne(padvances2);
		}

		while (padvances1) {
			to = FirstOne(padvances1);
			if (to < 56)
				*move++ = (to+8)|(to<<6)|(PAWN<<12);
			else {
				*move++=(to+8)|(to<<6)|(PAWN<<12)|(ROOK<<18);
				*move++=(to+8)|(to<<6)|(PAWN<<12)|(BISHOP<<18);
				*move++=(to+8)|(to<<6)|(PAWN<<12)|(KNIGHT<<18); 				
			}
			ClearFirstOne(padvances1);
		}

		target = And(white_pieces, rank_mask[RANK1]);
		pcapturesl = And(Shr(And(bPawns, mask_left_edge),9),target);
		pcapturesr = And(Shr(And(bPawns, mask_right_edge),7),target);

		while (pcapturesl) {
			to=FirstOne(pcapturesl);
			*move++=(to+9)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(ROOK<<18);
			*move++=(to+9)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(BISHOP<<18);
			*move++=(to+9)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(KNIGHT<<18);
			ClearFirstOne(pcapturesl);
		}
		while (pcapturesr) {
			to=FirstOne(pcapturesr);
			*move++=(to+7)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(ROOK<<18);
			*move++=(to+7)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(BISHOP<<18);
			*move++=(to+7)|(to<<6)|(PAWN<<12)|(PieceOnSquare(to)<<15)|(KNIGHT<<18);
			ClearFirstOne(pcapturesr);
		}

		promotions = And(Shr(And(bPawns, rank_mask[RANK2]),8),Compl(all));
		while (promotions){
			to = FirstOne(promotions);
			*move++ = (to+8)|(to<<6)|(PAWN<<12)|(KNIGHT<<18);
			*move++ = (to+8)|(to<<6)|(PAWN<<12)|(ROOK<<18);
			*move++ = (to+8)|(to<<6)|(PAWN<<12)|(BISHOP<<18);
			ClearFirstOne(promotions);
		}
	}

	return move;
}

extern int directions[64][64];


int* Board::GenerateCheckEvasions(int* move, int wtm){
	int checkers, checking_square;
	BITBOARD target, checksqs, moves, piecebd;
	int temp;
	int check_direction1 = 0, check_direction2 = 0;
	int from, to;
	int king_square;
	BITBOARD padvances1, padvances2, targetp, pcapturesl, pcapturesr;

	int* move0 = move;
	
	/* Determine the number or pieces attacking the king */
	if (wtm){
		king_square = white_king_sq;

		checksqs = And(AttacksTo(king_square, WHITE), black_pieces);
		checkers = BitCount(checksqs);
		if (checkers == 1){
			checking_square = FirstOne(checksqs);
			if (sq[checking_square] != -PAWN)
				check_direction1 = directions[checking_square][king_square];
			target = InterposeSquares(check_direction1, king_square, checking_square);
			target = Or(target, checksqs);
			target = Or(target, bKing);
		} else {
			target = bKing;
			checking_square = FirstOne(checksqs);
			if (sq[checking_square] != -PAWN)
				check_direction1 = directions[checking_square][king_square];
			Clear(checksqs, checking_square);
			checking_square=FirstOne(checksqs);
			if (sq[checking_square] != -PAWN)
				check_direction2 = directions[checking_square][king_square];
		}

		/* Produce a set of valid destination squares */
		
		// king moves
		
		from = king_square;
		moves = And(king_attacks[from], Compl(white_pieces));
		temp = from + (KING << 12);
		while (moves){
			to = FirstOne(moves);

			if (!Attacked(to, 1-wtm) && directions[from][to] != check_direction1 && directions[from][to] != check_direction2)
				*move++ = temp | (to << 6) | ((-PieceOnSquare(to))<<15);

			ClearFirstOne(moves);
		}

		if (checkers == 1) {

			// knight moves
			piecebd = wKnights;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(knight_attacks[from], target);
					temp = from + (KNIGHT << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | ((-PieceOnSquare(to))<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}

			// bishop moves
			piecebd = wBishops;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(AttacksBishop(from), target);
					temp = from + (BISHOP << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | ((-PieceOnSquare(to))<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}

			// rook moves
			piecebd = wRooks;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(AttacksRook(from), target);
					temp = from + (ROOK << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | ((-PieceOnSquare(to))<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}

			// queen moves
			piecebd = wQueens;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(AttacksQueen(from), target);
					temp = from + (QUEEN << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | ((-PieceOnSquare(to))<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}


			// pawn moves
			padvances1 = And(Shl(wPawns,8),Compl(all));
			padvances2 = And(Shl(And(padvances1, rank_mask[RANK3]),8),Compl(all));

			padvances1 = And(padvances1, target);
			padvances2 = And(padvances2, target);

			while (padvances2) {
				to = FirstOne(padvances2);
				if (!PinnedOnKing(wtm, to-16))
					*move++ = (to-16)|(to<<6)|(PAWN<<12);
				ClearFirstOne(padvances2);
			}

			while (padvances1) {
				to = FirstOne(padvances1);
				if (!PinnedOnKing(wtm, to-8)){
					if (to < 56)
						*move++ = (to-8)|(to<<6)|(PAWN<<12);
					else {
						*move++=(to-8)|(to<<6)|(PAWN<<12)|(QUEEN<<18);
						*move++=(to-8)|(to<<6)|(PAWN<<12)|(ROOK<<18);
						*move++=(to-8)|(to<<6)|(PAWN<<12)|(BISHOP<<18);
						*move++=(to-8)|(to<<6)|(PAWN<<12)|(KNIGHT<<18); 				
					}
				}
				ClearFirstOne(padvances1);
			}

			if (ep_square){
				targetp = Or(And(black_pieces, target), And(Shl(target, 8), set_mask[ep_square]));
			} else {
				targetp = And(black_pieces, target);
			}
			pcapturesl = And(Shl(And(wPawns, mask_left_edge),7),targetp);
			pcapturesr = And(Shl(And(wPawns, mask_right_edge),9),targetp);
	
			while (pcapturesl) {
				to=FirstOne(pcapturesl);
				if (!PinnedOnKing(wtm, to-7)){
					if (to < 56){
						if (PieceOnSquare(to)) // !ep ?
							*move++ = (to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15);
						else
							*move++ = (to-7)|(to<<6)|(PAWN<<12)|(PAWN<<15);
					} else {
						*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(QUEEN<<18);
						*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(ROOK<<18);
						*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(BISHOP<<18);
						*move++=(to-7)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(KNIGHT<<18);
					}
				}
				ClearFirstOne(pcapturesl);
			}
			while (pcapturesr) {
				to=FirstOne(pcapturesr);
				if (!PinnedOnKing(wtm, to-9)){
					if (to < 56){
						if (PieceOnSquare(to)) // !ep ?
							*move++ = (to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15);
						else
							*move++ = (to-9)|(to<<6)|(PAWN<<12)|(PAWN<<15);
					} else {
						*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(QUEEN<<18);
						*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(ROOK<<18);
						*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(BISHOP<<18);
						*move++=(to-9)|(to<<6)|(PAWN<<12)|((-PieceOnSquare(to))<<15)|(KNIGHT<<18);
					} 
				}
				ClearFirstOne(pcapturesr);
			}
		}
	
	} else {


		/* Determine the number or pieces attacking the king */
		king_square = black_king_sq;

		checksqs = And(AttacksTo(king_square, BLACK), white_pieces);
		checkers = BitCount(checksqs);
		if (checkers == 1){
			checking_square = FirstOne(checksqs);
			if (sq[checking_square] != PAWN)
				check_direction1 = directions[checking_square][king_square];
			target = InterposeSquares(check_direction1, king_square, checking_square);
			target = Or(target, checksqs);
			target = Or(target, wKing);
		} else {
			target = wKing;
			checking_square = FirstOne(checksqs);
			if (sq[checking_square] != PAWN)
				check_direction1 = directions[checking_square][king_square];
			Clear(checksqs, checking_square);
			checking_square=FirstOne(checksqs);
			if (sq[checking_square] != PAWN)
				check_direction2 = directions[checking_square][king_square];
		}

		/* Produce a set of valid destination squares */
		
		// king moves
		
		from = king_square;
		moves = And(king_attacks[from], Compl(black_pieces));
		temp = from + (KING << 12);
		while (moves){
			to = FirstOne(moves);

			if (!Attacked(to, 1-wtm) && directions[from][to] != check_direction1 && directions[from][to] != check_direction2)
				*move++ = temp | (to << 6) | (PieceOnSquare(to)<<15);

			ClearFirstOne(moves);
		}

		if (checkers == 1) {

			// knight moves
			piecebd = bKnights;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(knight_attacks[from], target);
					temp = from + (KNIGHT << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | (PieceOnSquare(to)<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}

			// bishop moves
			piecebd = bBishops;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(AttacksBishop(from), target);
					temp = from + (BISHOP << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | (PieceOnSquare(to)<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}

			// rook moves
			piecebd = bRooks;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(AttacksRook(from), target);
					temp = from + (ROOK << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | (PieceOnSquare(to)<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}

			// queen moves
			piecebd = bQueens;
			while (piecebd){
				from = FirstOne(piecebd);
				if (!PinnedOnKing(wtm, from)){
					moves = And(AttacksQueen(from), target);
					temp = from + (QUEEN << 12);
					while (moves){
						to = FirstOne(moves);
						*move++ = temp | (to << 6) | (PieceOnSquare(to)<<15);
						ClearFirstOne(moves);
					}
				}
				ClearFirstOne(piecebd);
			}


			// pawn moves
			padvances1 = And(Shr(bPawns,8),Compl(all));
			padvances2 = And(Shr(And(padvances1, rank_mask[RANK6]),8),Compl(all));

			padvances1 = And(padvances1, target);
			padvances2 = And(padvances2, target);

			while (padvances2) {
				to = FirstOne(padvances2);
				if (!PinnedOnKing(wtm, to+16))
					*move++ = (to+16)|(to<<6)|(PAWN<<12);
				ClearFirstOne(padvances2);
			}

			while (padvances1) {
				to = FirstOne(padvances1);
				if (!PinnedOnKing(wtm, to+8)){
					if (to > 7)
						*move++ = (to+8)|(to<<6)|(PAWN<<12);
					else {
						*move++=(to+8)|(to<<6)|(PAWN<<12)|(QUEEN<<18);
						*move++=(to+8)|(to<<6)|(PAWN<<12)|(ROOK<<18);
						*move++=(to+8)|(to<<6)|(PAWN<<12)|(BISHOP<<18);
						*move++=(to+8)|(to<<6)|(PAWN<<12)|(KNIGHT<<18); 				
					}
				}
				ClearFirstOne(padvances1);
			}

			if (ep_square){
				targetp = Or(And(white_pieces, target), And(Shr(target, 8), set_mask[ep_square]));
			} else {
				targetp = And(white_pieces, target);
			}
			pcapturesl = And(Shr(And(bPawns, mask_left_edge),9),targetp);
			pcapturesr = And(Shr(And(bPawns, mask_right_edge),7),targetp);

			while (pcapturesl) {
				to=FirstOne(pcapturesl);
				if (!PinnedOnKing(wtm, to+9)){
					if (to > 7){
						if (PieceOnSquare(to)) // !ep ?
							*move++ = (to+9)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15);
						else
							*move++ = (to+9)|(to<<6)|(PAWN<<12)|(PAWN<<15);
					} else {
						*move++=(to+9)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(QUEEN<<18);
						*move++=(to+9)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(ROOK<<18);
						*move++=(to+9)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(BISHOP<<18);
						*move++=(to+9)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(KNIGHT<<18);
					}
				}
				ClearFirstOne(pcapturesl);
			}
			while (pcapturesr) {
				to=FirstOne(pcapturesr);
				if (!PinnedOnKing(wtm, to+7)){
					if (to > 7){
						if (PieceOnSquare(to)) // !ep ?
							*move++ = (to+7)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15);
						else
							*move++ = (to+7)|(to<<6)|(PAWN<<12)|(PAWN<<15);
					} else {
						*move++=(to+7)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(QUEEN<<18);
						*move++=(to+7)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(ROOK<<18);
						*move++=(to+7)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(BISHOP<<18);
						*move++=(to+7)|(to<<6)|(PAWN<<12)|((PieceOnSquare(to))<<15)|(KNIGHT<<18);
					} 
				}
				ClearFirstOne(pcapturesr);
			}
		}
	}

	#if VERIFY_CHECK_EVASIONS

	static int m_list[256];
	int* last = GenerateAllValidMoves(m_list, wtm);	

	int identical = 1;
	
	if ((last-m_list) == (move - move0)){
		int* m_i, *m_j;
		int found;

		for (m_i = move0; m_i < move; m_i++){
			found = 0;
			for (m_j = m_list; m_j < last; m_j++)
				if (*m_i == *m_j){ found = 1;break;	}

			if (!found) { identical = 0; break; }
		}
		for (m_i = m_list; m_i < last; m_i++){
			found = 0;
			for (m_j = move0; m_j < move; m_j++)
				if (*m_i == *m_j){ found = 1;break;	}

			if (!found) { identical = 0; break; }
		}
	}
	

	if (((last-m_list) != (move - move0)) || (!identical)){
		board.Print();
		cout << "board epd : " << board.PrintEPD() << endl;

		int*m;
		cout << "movelist1 : "; for (m = move0; m < move; m++) PrintMove(*m); cout << endl;
		cout << "movelist2 : "; for (m = m_list; m < last; m++) PrintMove(*m); cout << endl;

		cout << "Target : " << endl;
		PrintBitboard(target);

		targetp = Or(And(white_pieces, target), And(Shr(target, 8), set_mask[ep_square]));

		cout << "Targetp : " << endl;
		PrintBitboard(targetp);

		cout << "ep_square = " << ep_square << endl;


		//padvances2 = And(Shl(And(padvances1, rank_mask[RANK3]),8),Compl(all));
		//padvances2 = And(padvances2, target);

		//padvances2 = 

		//PrintBitboard(padvances2);

	}

	Assert((last-m_list) == (move - move0));

	#endif

	return move;
}
