#include <stdio.h>
#include "debug.h"
#include <string.h>
#include "stdlib.h"
#include <iostream>
#include "board.h"
#include "move.h"
#include "tree.h"

using namespace std;


//#define VERBOSE

int MAX_MOVES_PER_GAME_TO_PARSE = 10; 


const int TEMP_BUF_SIZE = 16384;

class TempFile {
public:
	FILE* fp;

	int count;

	struct {
		HASHCODE code;
		char result;
	} buf[TEMP_BUF_SIZE];

	void Dump();

	void Open(char* filename);
	void Close();
	void Write(HASHCODE parent, HASHCODE child, char* result);

} temp_file;

void TempFile::Open(char* filename){
	fp = fopen(filename, "wb");
	Assert(filename);

	count = 0;
}

void TempFile::Dump(){
	if (count)
		fwrite(buf, sizeof(buf[0]), count, fp);
	count = 0;
}

void TempFile::Close(){
	Dump();

	fclose(fp);
}

void TempFile::Write(HASHCODE parent, HASHCODE child, char* result){
	BITBOARD hashcode = Shl(Shr(parent,48),48) + Shr(Shl(child,16),16);

	char res = 0;

	if (strstr(result, "1-0"))
		res = 1;
	else if (strstr(result, "0-1"))
		res = -1;
	else
		res = 0;

	if (count >= TEMP_BUF_SIZE)
		Dump();

	buf[count].code = hashcode;
	buf[count].result = res;
	count++;

	//fwrite(&hashcode, 1, 8, fp);
	//fwrite(&res, 1, 1, fp);
}






int FindMove(int* m, int* last, int from, int to, int piece){
	int* start = m;

	for ( ; m < last; m++){
		int f = From(*m);
		int t = To(*m);
		int p = Piece(*m);

		if (t == to && p == piece){
			if (from >= 0 && from != f)
				continue;
			return *m;
		}
	}

	board.Print();
	cout << "Find move : can't find " << from << endl << to << endl << piece << endl;
	PrintMoveList(start);
	cout << endl;

	return 0;
}

// finds piece moving from specified rank to destination
int FindRankMove(int* m, int* last, int rank, int to, int piece){
	int* start = m;

	for ( ; m < last; m++){
		int f = From(*m);
		int t = To(*m);
		int p = Piece(*m);

		if (t == to && p == piece && Rank(f) == rank)
			return *m;
	}
	
	board.Print();
	cout << "Find move rank : can't find " << rank << endl << to << endl << piece << endl;
	PrintMoveList(start);
	cout << endl;

	return 0;
}

// finds piece moving from specified rank to destination
int FindFileMove(int* m, int* last, int file, int to, int piece){
	int* start = m;

	for ( ; m < last; m++){
		int f = From(*m);
		int t = To(*m);
		int p = Piece(*m);

		if (t == to && p == piece && File(f) == file)
			return *m;
	}

	board.Print();
	cout << "Find move file : can't find " << file << endl << to << endl << piece << endl;
	PrintMoveList(start);
	cout << endl;

	return 0;
}

int TestMove(char* move, int wtm){
	static int move_list0[512], move_list[512];
	int *last, *last0;

	cout << "Test move : [" << move << "]" << endl;
	
	char* s = move;

	last0 = board.GenerateCaptures(move_list0,wtm);
	last0 = board.GenerateNonCaptures(last0, wtm);
	*last0 = 0;

	//cout << "  move list :"; PrintMoveList(move_list); cout << endl;

	int piece,to;

	switch(s[0]){
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
		piece = PAWN;
		break;
	case 'K':
		piece = KING;
		break;
	case 'N':
		piece = KNIGHT;
		break;
	case 'B':
		piece = BISHOP;
		break;
	case 'R':
		piece = ROOK;
		break;
	case 'Q':
		piece = QUEEN;
		break;
	case 'O': // castle
	case 'o':
		piece = KING;
		break;
	default:
		if (!strstr(move, "1-0") && !strstr(move, "0-1") && !strstr(move, "1/2-1/2"))
			cout << "move " << move << " is not valid" << endl;
		return 0;	// non valid move
	}

	SaveBoardInfo(0);

	// remove moves that leaves the king in check
	last = move_list;
	for (int* m = move_list0; m < last0; m++){
		// verify if the moved piece doesn't leave the king in check
		if (Piece(*m) == piece){
			board.Make(*m, 0, wtm);
			if (!Check(wtm))
				*last++ = *m;
			board.Unmake(*m, 0, wtm);
		}
	}
	*last = 0;


	if (s[0] == 'o' || s[0] == 'O')
	if (!stricmp(s, "o-o-o")){
		if (wtm) return FindMove(move_list, last, E1, C1, KING);
			else return FindMove(move_list, last, E8, C8, KING);
	} else if (!stricmp(s, "o-o")){
		if (wtm) return FindMove(move_list, last, E1, G1, KING);
			else return FindMove(move_list, last, E8, G8, KING);
	} else {
		Assert(false);
	}

	int len = strlen(s);

	#define IsLetter(x) (((x)>='a')&&((x)<='h'))
	#define IsDigit(x) (((x)>='1')&&((x)<='8'))

	if (piece == PAWN){
		// five forms 
		//	['a'..'h']['1'..'8'], last 2 aren't used for now, all promotions are done to queen
		//	['a'..'h']['x']['a'..'h']['1'..'8']
		//  ['a'..'h']['a'..'h']['1'..'8'] for captures
		//	['a'..'h']['1'..'8']['Q','R','N','B'] for promotions
		//	['a'..'h']['x']['a'..'h']['1'..'8']['Q','R','N','B'] for promotions
		if (IsLetter(s[0]) && IsDigit(s[1])){
			to = Square(s[1]-'1', s[0]-'a');
			return FindMove(move_list, last, -1, to, PAWN);
		} else {
			if (s[1] == 'x'){
				Assert(IsLetter(s[2]));
				Assert(IsDigit(s[3]));
				return FindFileMove(move_list, last, s[0]-'a',Square(s[3]-'1', s[2]-'a'),piece);
			} else {
				Assert(IsLetter(s[1]));
				Assert(IsDigit(s[2]));
				return FindFileMove(move_list, last, s[0]-'a',Square(s[2]-'1', s[1]-'a'),piece);
			}
		}

		// bug : promotions to anything else than queen aren't handled right
	}

	s++;
	if (*s == 'x') s++;

	// the following situations are possible :
	//	['a'..'h']['a'..'h']['1'..8']
	//	['1'..'8']['a'..'h']['1'..8']
	//	['a'..'h']['1'..'8']
	//  ['a'..'h']['x']['a'..'h']['1'..8']
	//	['1'..'8']['x']['a'..'h']['1'..8']

	if (s[1] == 'x'){
		if (IsDigit(s[0])){
			Assert(IsLetter(s[2]));
			Assert(IsDigit(s[3]));
			return FindRankMove(move_list, last, s[0]-'1',Square(s[3]-'1', s[2]-'a'),piece);
		} else {
			Assert(IsDigit(s[3]));
			Assert(IsLetter(s[0]));
			Assert(IsLetter(s[2]));
			return FindFileMove(move_list, last, s[0]-'a',Square(s[3]-'1', s[2]-'a'),piece);
		}
	}
	if (IsDigit(s[0])){
		//cout << "debug : " << move << endl;
		Assert(IsLetter(s[1]));
		Assert(IsDigit(s[2]));
		return FindRankMove(move_list, last, s[0]-'1',Square(s[2]-'1', s[1]-'a'),piece);
	} else {
		Assert(IsLetter(s[0]))
		if (IsLetter(s[1])){
			Assert(IsDigit(s[2]));
			return FindFileMove(move_list, last, s[0]-'a',Square(s[2]-'1', s[1]-'a'),piece);
		} else {
			Assert(IsDigit(s[1]));
			return FindMove(move_list, last, -1,Square(s[1]-'1', s[0]-'a'),piece);
		}
	}

	board.Print();
	cout << "Not found move is " << move << endl;
	cout << "Move list for this position is : ";
	PrintMoveList(move_list);

	return 0;
}


// returns the number of valid moves
int ParsePgn(char* moves, int* move_list, int max_moves, char* result) {
	char buf1[10],buf2[10], str[10];
	char* p;

	board.Init("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w");
	board.castle_flags = WHITE_LCASTLE | WHITE_SCASTLE | BLACK_LCASTLE | BLACK_SCASTLE;
	int wtm = 1;

	HASHCODE parent = 0;
	HASHCODE child = board.hashcode;

	int move;

	SaveBoardInfo(0);

	for (int i = 0; i < max_moves; i++){
		itoa(i+1, str,10);
		strcat(str, ".");
		char* s = strstr(moves, str);

		if (!s) break;

		s+= strlen(str);

		// strip whitespaces
		while (*s <= 32)
			s++;

		// copy 1st move
		for (p = buf1; p-buf1 < 10 && *s != 0 && *s > 32;)
			*p++ = *s++;
		*p++ = 0;

#ifdef VERBOSE	
		cout << buf1 << endl;
#endif
		if ((move = TestMove(buf1, 1))){
			*move_list++ = move;
			board.Make(move, 0, 1);

			parent = child;
			child = board.hashcode;
			//ob.Add(parent, child, result);
			temp_file.Write(parent, child, result);

			/*
			HASHCODE hc = Shl(Shr(parent,48),48) + Shr(Shl(child,16),16);
			cout << "parent = "; Print(parent); cout << endl;
			cout << "child = "; Print(child); cout << endl;
			Print(hc); cout << endl;
			*/

		} else {
			*move_list++ = 0;
			if (!strstr(buf1, "1-0") && !strstr(buf1, "0-1") && !strstr(buf1, "1/2-1/2"))
				cout << "ParsePgn : move " << buf1 << " couldn't be converted" << endl;
			return 0;
		}

		// strip whitespaces
		while (*s <= 32)
			s++;

		// copy 2nd move
		for (p = buf2; p-buf2 < 10 && *s != 0 && *s > 32;)
			*p++ = *s++;

		*p++ = 0;

#ifdef VERBOSE	
		cout << buf2;
#endif

		if ((move = TestMove(buf2, 0))){
			*move_list++ = move;
			board.Make(move, 0, 0);

			parent = child;
			child = board.hashcode;
			//ob.Add(parent, child, result);
			temp_file.Write(parent, child, result);

			/*
			cout << "parent = "; Print(parent); cout << endl;
			cout << "child = "; Print(child); cout << endl;
			HASHCODE hc = Shl(Shr(parent,48),48) + Shr(Shl(child,16),16);
			Print(hc); cout << endl;
			*/

		} else {
			*move_list++ = 0;
			if (!strstr(buf2, "1-0") && !strstr(buf2, "0-1") && !strstr(buf2, "1/2-1/2"))
				cout << "ParsePgn : move " << buf2 << " couldn't be converted" << endl;
			return 0;
		}

#ifdef VERBOSE	
		cout << endl;
#endif

		//cout << i+1 << ". \"" << buf1 << "\" \"" << buf2 << "\"" << endl;
	}


	*move_list++ = 0;

	return 0;
}

typedef enum {OUTSIDE, IN_HEADER, IGNORE, MOVE_LIST} states;

void ParseHeaderLine(char* line, char* result){

#if 0
	// returns the content from line like [Name "xxx"]
	line = strstr(line, "\"");
	if (!line) 
		result[0] = 0;
	else {
		int counter = strlen(result);
		Assert(counter < 256);
		line++;
		for ( ; *line != 0 && *line != '\"'; line++, counter++){
			Assert(counter < 256);
			*result++ = *line;
		}
		*result = 0;
	}
#else
	result[0] = 0;
#endif
}

static char* p_buf;
static int move_list[4096];
static char buf[16384];
static int state = OUTSIDE;
static char result[256];

static int header_level= 0;

int processed_positions = 0;

void ProcessMovelist(){
	
#ifdef VERBOSE	
	cout << "Parsed the move list : " << buf << endl;
#endif

	char* result;
	result = strstr(buf, "1/2-1/2");
	if (!result)
		result = strstr(buf, "1-0");
	if (!result)
		result = strstr(buf, "0-1");
				
	if (result){
#ifdef VERBOSE	
		cout << result << endl;
#endif
		ParsePgn(buf, move_list, MAX_MOVES_PER_GAME_TO_PARSE, result);

#ifdef VERBOSE	
		cout << endl << "moves : ";
		PrintMoveList(move_list);
		cout << result << endl;
#endif
	} else {
#ifdef VERBOSE	
		cout << "Can't identify result" << endl;
#endif
	}

	processed_positions++;
	if (processed_positions % 50 == 0) 
		cout << processed_positions << " so far" << endl;
}


void ProcessChar(char c){
	switch (state){
	case OUTSIDE:
		if (c == '['){
			header_level = 1;
			state = IN_HEADER;
			p_buf = buf;
			*p_buf = c; p_buf++; *p_buf = 0;
		}else if (c > 32){
			state = MOVE_LIST;
			p_buf = buf;
			*p_buf = c; p_buf++; *p_buf = 0;
		}
		break;
	case IN_HEADER:
		if (c == '\"')
			state = IGNORE;
		if (c == '[')
			header_level++;
		Assert(strlen(buf) < 16300);
		*p_buf = c; p_buf++; *p_buf = 0;
		if (c == ']'){
			header_level--;
			if (header_level == 0){
				ParseHeaderLine(buf, result);
				buf[0] = 0;
				state = OUTSIDE;
			}
		} 
		break;
	case IGNORE:
		if (c == '\"')
			state = IN_HEADER;
		else {
			if (strlen(buf) >= 16300){
				cout << "buf = {\n" << buf << "\n}\n";
			}
			Assert(strlen(buf) < 16300);
			*p_buf = c; p_buf++; *p_buf = 0;
		}
		break;
	case MOVE_LIST:
		if (c == '['){
			header_level = 1;
			ProcessMovelist();
			state = IN_HEADER;
			Assert(p_buf-buf < 16000);
			p_buf = buf;
			*p_buf = c; p_buf++; *p_buf = 0;
		} else {
			Assert(p_buf-buf < 16000);
			*p_buf = c; p_buf++; *p_buf = 0;
		}
		break;
	default:
		Assert(false);
	}

}


void ParsePgn(char* filename, char* temp_filename, int max_book_moves){
	char line[1024];

	MAX_MOVES_PER_GAME_TO_PARSE = max_book_moves;

	temp_file.Open(temp_filename);

	memset(move_list, 0, sizeof(move_list));
	buf[0] = 0;

	FILE* fp = fopen(filename, "rt");
	if (!fp) return;

	while (!feof(fp)){
		fgets(line, 1024, fp);

		cout << line;

		if (strstr(line, "[") && strstr(line, "]")){ // ignore []
			ProcessChar('[');
			ProcessChar('T');
			ProcessChar(']');	
		} else {
			// remove occurences like \" which will make my parser go crazy
			for (char* p = line; *p; p++)
				if ((*p == '\\') && (*(p+1) == '\"') && (*(p+2) != ']')){
					*(p+1) = 'o';
				}

			for (char* m = line; *m; m++)
				ProcessChar(*m);
		}
	}

	if (state == MOVE_LIST)
		ProcessMovelist();

	fclose(fp);

	temp_file.Close();
}

